/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.16 from the
 * contents of mRuby.xs. Do not edit this file, edit mRuby.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "lib/mRuby.xs"
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "../ppport.h"

#include "mruby.h"
#include "mruby/array.h"
#include "mruby/class.h"
#include "mruby/compile.h"
#include "mruby/hash.h"
#include "mruby/khash.h"
#include "mruby/proc.h"
#include "mruby/string.h"
#include "mruby/variable.h"

KHASH_DECLARE(ht, mrb_value, mrb_value, 1);
/* KHASH_DEFINE (ht, mrb_value, mrb_value, 1, mrb_hash_ht_hash_func, mrb_hash_ht_hash_equal); */

static
SV * mrb_value2sv(mrb_value v) {
    switch (mrb_type(v)) {
    case MRB_TT_FALSE:
        return &PL_sv_undef;
    case MRB_TT_FIXNUM:
        return newSViv(mrb_fixnum(v));
    case MRB_TT_STRING:
        return newSVpv((char*)RSTRING_PTR(v), (STRLEN)RSTRING_LEN(v));
    case MRB_TT_HASH: {
        HV * ret = newHV();
        khash_t(ht) * h = RHASH_TBL(v);
        khiter_t k;

        if (!h) { abort(); }
        for (k = kh_begin(h); k != kh_end(h); k++) {
            if (kh_exist(h, k)) {
                mrb_value kk = kh_key(h,k);
                mrb_value vv = kh_value(h,k);

                STRLEN key_len;
                char *key = SvPV(mrb_value2sv(kk), key_len);
                hv_store(ret, key, key_len, mrb_value2sv(vv), 0);
            }
        }

        return newRV_noinc((SV*)ret);
    }
    case MRB_TT_ARRAY: {
        int len = RARRAY_LEN(v);
        AV * ret = newAV();
        int i;
        mrb_value *ptr = RARRAY_PTR(v);
        for (i=0; i<len; i++) {
            av_push(ret, mrb_value2sv(ptr[i]));
        }
        return newRV_noinc((SV*)ret);
    }
    default:
        croak("This type of ruby value is not supported yet: %d", mrb_type(v));
    }
    abort();
}

#line 74 "lib/mRuby.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(pTHX_ const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(pTHX_ const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
            Perl_croak(aTHX_ "Usage: %s::%s(%s)", hvname, gvname, params);
        else
            Perl_croak(aTHX_ "Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
        Perl_croak(aTHX_ "Usage: CODE(0x%"UVxf")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#ifdef PERL_IMPLICIT_CONTEXT
#define croak_xs_usage(a,b)    S_croak_xs_usage(aTHX_ a,b)
#else
#define croak_xs_usage        S_croak_xs_usage
#endif

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#line 216 "lib/mRuby.c"

XS_EUPXS(XS_mRuby__State_new); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_mRuby__State_new)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "klass");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	const char *	klass = (const char *)SvPV_nolen(ST(0))
;
#line 71 "lib/mRuby.xs"
        mrb_state* mrb = mrb_open();
        XPUSHs(sv_bless(newRV_noinc(sv_2mortal(newSViv(PTR2IV(mrb)))), gv_stashpv(klass, TRUE)));
        XSRETURN(1);
#line 233 "lib/mRuby.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_mRuby__State_parse_string); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_mRuby__State_parse_string)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "mrb, src");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	mrb_state *	mrb;
	const char *	src = (const char *)SvPV_nolen(ST(1))
;

    if ( sv_isobject( ST(0) ) && ( SvTYPE( SvRV( ST(0) ) ) == SVt_PVMG ) )
    {
        mrb = (mrb_state *)SvIV( (SV*)SvRV( ST(0) ) );
    }
    else
    {
        warn( "mRuby::State::parse_string() -- mrb is not a blessed SV reference" );
        XSRETURN_UNDEF;
    }
;
#line 78 "lib/mRuby.xs"
        struct mrb_parser_state* st = mrb_parse_string(mrb, src);
        XPUSHs(sv_bless(newRV_noinc(sv_2mortal(newSViv(PTR2IV(st)))), gv_stashpv("mRuby::ParserState", TRUE)));
        XSRETURN(1);
#line 267 "lib/mRuby.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_mRuby__State_generate_code); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_mRuby__State_generate_code)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "mrb, st");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	mrb_state *	mrb;
	struct mrb_parser_state*	st;

    if ( sv_isobject( ST(0) ) && ( SvTYPE( SvRV( ST(0) ) ) == SVt_PVMG ) )
    {
        mrb = (mrb_state *)SvIV( (SV*)SvRV( ST(0) ) );
    }
    else
    {
        warn( "mRuby::State::generate_code() -- mrb is not a blessed SV reference" );
        XSRETURN_UNDEF;
    }
;

    if ( sv_isobject( ST(1) ) && ( SvTYPE( SvRV( ST(1) ) ) == SVt_PVMG ) )
    {
        st = (struct mrb_parser_state *)SvIV( (SV*)SvRV( ST(1) ) );
    }
    else
    {
        warn( "mRuby::State::generate_code() -- st is not a blessed SV reference" );
        XSRETURN_UNDEF;
    }
;
#line 85 "lib/mRuby.xs"
        int n = mrb_generate_code(mrb, st->tree);
        XSRETURN_IV(n);
#line 310 "lib/mRuby.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_mRuby__State_proc_new); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_mRuby__State_proc_new)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "mrb, n");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	mrb_state *	mrb;
	int	n = (int)SvIV(ST(1))
;

    if ( sv_isobject( ST(0) ) && ( SvTYPE( SvRV( ST(0) ) ) == SVt_PVMG ) )
    {
        mrb = (mrb_state *)SvIV( (SV*)SvRV( ST(0) ) );
    }
    else
    {
        warn( "mRuby::State::proc_new() -- mrb is not a blessed SV reference" );
        XSRETURN_UNDEF;
    }
;
#line 91 "lib/mRuby.xs"
        struct RProc * proc = mrb_proc_new(mrb, mrb->irep[n]);
        XPUSHs(sv_bless(newRV_noinc(sv_2mortal(newSViv(PTR2IV(proc)))), gv_stashpv("mRuby::RProc", TRUE)));
        XSRETURN(1);
#line 344 "lib/mRuby.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_mRuby__State_run); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_mRuby__State_run)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "mrb, proc, val");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	mrb_state *	mrb;
	struct RProc*	proc;
	SV *	val = ST(2)
;

    if ( sv_isobject( ST(0) ) && ( SvTYPE( SvRV( ST(0) ) ) == SVt_PVMG ) )
    {
        mrb = (mrb_state *)SvIV( (SV*)SvRV( ST(0) ) );
    }
    else
    {
        warn( "mRuby::State::run() -- mrb is not a blessed SV reference" );
        XSRETURN_UNDEF;
    }
;

    if ( sv_isobject( ST(1) ) && ( SvTYPE( SvRV( ST(1) ) ) == SVt_PVMG ) )
    {
        proc = (struct RProc *)SvIV( (SV*)SvRV( ST(1) ) );
    }
    else
    {
        warn( "mRuby::State::run() -- proc is not a blessed SV reference" );
        XSRETURN_UNDEF;
    }
;
#line 98 "lib/mRuby.xs"
        mrb_value ret = mrb_run(mrb, proc, mrb_nil_value());
        mXPUSHs(mrb_value2sv(ret));
        XSRETURN(1);
#line 390 "lib/mRuby.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_mRuby__ParserState_pool_close); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_mRuby__ParserState_pool_close)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "st");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	struct mrb_parser_state*	st;

    if ( sv_isobject( ST(0) ) && ( SvTYPE( SvRV( ST(0) ) ) == SVt_PVMG ) )
    {
        st = (struct mrb_parser_state *)SvIV( (SV*)SvRV( ST(0) ) );
    }
    else
    {
        warn( "mRuby::ParserState::pool_close() -- st is not a blessed SV reference" );
        XSRETURN_UNDEF;
    }
;
#line 107 "lib/mRuby.xs"
        mrb_pool_close(st->pool);
        XSRETURN_UNDEF;
#line 421 "lib/mRuby.c"
	PUTBACK;
	return;
    }
}

#ifdef __cplusplus
extern "C"
#endif
XS_EXTERNAL(boot_mRuby); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_mRuby)
{
    dVAR; dXSARGS;
#if (PERL_REVISION == 5 && PERL_VERSION < 9)
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#endif
    XS_VERSION_BOOTCHECK;

        newXS("mRuby::State::new", XS_mRuby__State_new, file);
        newXS("mRuby::State::parse_string", XS_mRuby__State_parse_string, file);
        newXS("mRuby::State::generate_code", XS_mRuby__State_generate_code, file);
        newXS("mRuby::State::proc_new", XS_mRuby__State_proc_new, file);
        newXS("mRuby::State::run", XS_mRuby__State_run, file);
        newXS("mRuby::ParserState::pool_close", XS_mRuby__ParserState_pool_close, file);
#if (PERL_REVISION == 5 && PERL_VERSION >= 9)
  if (PL_unitcheckav)
       call_list(PL_scopestack_ix, PL_unitcheckav);
#endif
    XSRETURN_YES;
}

