package mRuby;
use strict;
use warnings;
use 5.008008;
our $VERSION = '0.01';

use XSLoader;

XSLoader::load(__PACKAGE__, $VERSION);

1;
__END__

=encoding utf8

=head1 NAME

mRuby - A module for you

=head1 SYNOPSIS

  use mRuby;

=head1 DESCRIPTION

mRuby is

=head1 AUTHOR

Tokuhiro Matsuno E<lt>tokuhirom AAJKLFJEF@ GMAIL COME<gt>

=head1 SEE ALSO

=head1 LICENSE

Copyright (C) Tokuhiro Matsuno

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut
